@@ -1,5 +1,5 @@
 /******************************************************************************
- * Copyright 1994-2002,2003 by Thomas E. Dickey                               *
+ * Copyright 1994-2004,2005 by Thomas E. Dickey                               *
  * All Rights Reserved.                                                       *
  *                                                                            *
  * Permission to use, copy, modify, and distribute this software and its      *
@@ -20,7 +20,7 @@
  ******************************************************************************/
 
 #ifndef	NO_IDENT
-static char *Id = "$Id: diffstat.c,v 1.34 2003/11/09 18:45:00 tom Exp $";
+static const char *Id = "$Id: diffstat.c,v 1.39 2005/04/10 14:55:15 tom Exp $";
 #endif
 
 /*
@@ -28,6 +28,24 @@ static char *Id = "$Id: diffstat.c,v 1.34 2003/11/09 18:45:00 tom Exp $";
  * Author:	T.E.Dickey
  * Created:	02 Feb 1992
  * Modified:
+ *		10 Apr 2005, change order of merging and prefix-stripping so
+ *			     stripping all prefixes, e.g., with -p9, will be
+ *			     sorted as expected (Patch by Jean Delvare
+ *			     <khali@linux-fr.org>).
+ *		10 Jan 2005, add support for '--help' and '--version' (Patch
+ *			     by Eric Blake <ebb9@byu.net>.)
+ *		16 Dec 2004, fix a different case for data beginning with "--"
+ *			     which was treated as a header line.
+ *		14 Dec 2004, Fix allocation problems.  Open files in binary
+ *			     mode for reading.  Getopt returns -1, not
+ *			     necessarily EOF.  Add const where useful.  Use
+ *			     NO_IDENT where necessary.  malloc() comes from
+ *			     <stdlib.h> in standard systems (Patch by Eric
+ *			     Blake <ebb9@byu.net>.)
+ *		08 Nov 2004, minor fix for resync of unified diffs checks for
+ *			     range (line beginning with '@' without header
+ *			     lines (successive lines beginning with "---" and
+ *			     "+++").  Fix a few problems reported by valgrind.
  *		09 Nov 2003, modify check for lines beginning with '-' or '+'
  *			     to treat only "---" in old-style diffs as a
  *			     special case.
@@ -117,8 +135,6 @@ extern int isatty();
 
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
-#else
-extern char *malloc();
 #endif
 
 #ifdef HAVE_GETOPT_H
@@ -191,12 +207,22 @@ static long plot_scale;		/* the effective scale (1:maximum) */
 /******************************************************************************/
 
 static void
-failed(char *s)
+failed(const char *s)
 {
     perror(s);
     exit(EXIT_FAILURE);
 }
 
+/* malloc wrapper that never returns NULL */
+static void *
+xmalloc(size_t s)
+{
+    void *p;
+    if ((p = malloc(s)) == NULL)
+	failed("malloc");
+    return p;
+}
+
 static void
 blip(int c)
 {
@@ -207,30 +233,45 @@ blip(int c)
 }
 
 static char *
-new_string(char *s)
+new_string(const char *s)
 {
-    return strcpy((char *) malloc((unsigned) (strlen(s) + 1)), s);
+    return strcpy((char *) xmalloc((size_t) (strlen(s) + 1)), s);
 }
 
 static DATA *
-new_data(char *name)
+new_data(const char *name)
 {
     DATA *p, *q, *r;
+    int base = 0;
 
     TRACE(("new_data(%s)\n", name));
 
+    /* Compute the base offset if the prefix option is used */
+    if (prefix_opt >= 0) {
+	int n;
+
+	for (n = prefix_opt; n > 0; n--) {
+	    char *s = strchr(name + base, PATHSEP);
+	    if (s == 0 || *++s == EOS)
+		break;
+	    base = s - name;
+	}
+	TRACE(("base set to %d\n", base));
+    }
+
     /* Insert into sorted list (usually sorted).  If we are not sorting or
      * merging names, we fall off the end and link the new entry to the end of
-     * the list.
+     * the list.  If the prefix option is used, the prefix is ignored by the
+     * merge and sort operations.
      */
     for (p = all_data, q = 0; p != 0; q = p, p = p->link) {
-	int cmp = strcmp(p->name, name);
+	int cmp = strcmp(p->name + p->base, name + base);
 	if (merge_names && (cmp == 0))
 	    return p;
 	if (sort_names && (cmp > 0))
 	    break;
     }
-    r = (DATA *) malloc(sizeof(DATA));
+    r = (DATA *) xmalloc(sizeof(DATA));
     if (q != 0)
 	q->link = r;
     else
@@ -238,7 +279,7 @@ new_data(char *name)
 
     r->link = p;
     r->name = new_string(name);
-    r->base = 0;
+    r->base = base;
     r->cmt = Normal;
     r->ins = 0;
     r->del = 0;
@@ -248,10 +289,9 @@ new_data(char *name)
 }
 
 /*
- * Remove a unneeded data item from the linked list.  Don't free the name,
- * since we may want it in another context.
+ * Remove a unneeded data item from the linked list.  Free the name as well.
  */
-static void
+static int
 delink(DATA * data)
 {
     DATA *p, *q;
@@ -264,15 +304,17 @@ delink(DATA * data)
 		q->link = p->link;
 	    else
 		all_data = p->link;
+	    free(p->name);
 	    free(p);
-	    return;
+	    return 1;
 	}
     }
+    return 0;
 }
 
 /* like strncmp, but without the 3rd argument */
 static int
-match(char *s, char *p)
+match(const char *s, const char *p)
 {
     int ok = 0;
 
@@ -283,12 +325,16 @@ match(char *s, char *p)
 	}
 	if (*s++ != *p++)
 	    break;
+	if (*s == EOS && *p == EOS) {
+	    ok = 1;
+	    break;
+	}
     }
     return ok;
 }
 
 static int
-version_num(char *s)
+version_num(const char *s)
 {
     int main_ver, sub_ver;
     char temp[2];
@@ -299,7 +345,7 @@ version_num(char *s)
  * Check for a range of line-numbers, used in editing scripts.
  */
 static int
-edit_range(char *s)
+edit_range(const char *s)
 {
     int first, last;
     char temp[2];
@@ -313,20 +359,22 @@ edit_range(char *s)
  * shows that the numbers are a line-number followed by a count.
  */
 static int
-decode_range(char *s, int *first, int *second)
+decode_range(const char *s, int *first, int *second)
 {
     char check;
     if (sscanf(s, "%d,%d%c", first, second, &check) == 2) {
+	TRACE(("decode_range #1 first=%d, second=%d\n", *first, *second));
 	return 1;
     } else if (sscanf(s, "%d%c", first, &check) == 1) {
 	*second = *first;	/* diffutils 2.7 does this */
+	TRACE(("decode_range #2 first=%d, second=%d\n", *first, *second));
 	return 1;
     }
     return 0;
 }
 
 static int
-HadDiffs(DATA * data)
+HadDiffs(const DATA * data)
 {
     return data->ins != 0
 	|| data->del != 0
@@ -338,7 +386,7 @@ HadDiffs(DATA * data)
  * If the given path is not one of the "ignore" paths, then return true.
  */
 static int
-can_be_merged(char *path)
+can_be_merged(const char *path)
 {
     if (strcmp(path, "")
 	&& strcmp(path, "/dev/null")
@@ -348,7 +396,7 @@ can_be_merged(char *path)
 }
 
 static int
-is_leaf(char *leaf, char *path)
+is_leaf(const char *leaf, const char *path)
 {
     char *s;
 
@@ -360,14 +408,15 @@ is_leaf(char *leaf, char *path)
 }
 
 static char *
-do_merging(DATA * data, char *path)
+do_merging(DATA * data, char *path, int *freed)
 {
     TRACE(("do_merging(%s,%s) diffs:%d\n", data->name, path, HadDiffs(data)));
 
+    *freed = 0;
     if (!HadDiffs(data)) {	/* the data was the first of 2 markers */
 	if (is_leaf(data->name, path)) {
 	    TRACE(("is_leaf: %s vs %s\n", data->name, path));
-	    delink(data);
+	    *freed = delink(data);
 	} else if (can_be_merged(data->name)
 		   && can_be_merged(path)) {
 	    size_t len1 = strlen(data->name);
@@ -405,7 +454,7 @@ do_merging(DATA * data, char *path)
 		&& diff)
 		path += len2 - matched + 1;
 
-	    delink(data);
+	    *freed = delink(data);
 	    TRACE(("merge @%d, prefix_opt=%d matched=%d diff=%d\n",
 		   __LINE__, prefix_opt, matched, diff));
 	} else if (!can_be_merged(path)) {
@@ -414,7 +463,7 @@ do_merging(DATA * data, char *path)
 	    path = data->name;
 	} else {
 	    TRACE(("merge @%d\n", __LINE__));
-	    delink(data);
+	    *freed = delink(data);
 	}
     } else if (!can_be_merged(path)) {
 	path = data->name;
@@ -424,7 +473,7 @@ do_merging(DATA * data, char *path)
 }
 
 static int
-begin_data(DATA * p)
+begin_data(const DATA * p)
 {
     if (!can_be_merged(p->name)
 	&& strchr(p->name, PATHSEP) != 0) {
@@ -465,9 +514,12 @@ static void
 dequote(char *s)
 {
     int len = strlen(s);
+    int n;
 
     if (*s == SQUOTE && len > 2 && s[len - 1] == SQUOTE) {
-	strcpy(s, s + 1);
+	for (n = 0; (s[n] = s[n + 1]) != EOS; ++n) {
+	    ;
+	}
 	s[len - 2] = EOS;
     }
 }
@@ -478,8 +530,7 @@ dequote(char *s)
 static void
 fixed_buffer(char **buffer, size_t want)
 {
-    if ((*buffer = malloc(want)) == 0)
-	failed("malloc");
+    *buffer = xmalloc(want);
 }
 
 /*
@@ -506,7 +557,7 @@ get_line(char **buffer, size_t *have, FILE *fp)
 	if (used + 2 > *have) {
 	    adjust_buffer(buffer, *have *= 2);
 	}
-	(*buffer)[used++] = ch;
+	(*buffer)[used++] = (char) ch;
 	if (ch == '\n')
 	    break;
     }
@@ -519,7 +570,9 @@ get_line(char **buffer, size_t *have, FILE *fp)
 static void
 do_file(FILE *fp)
 {
-    DATA dummy, *that = &dummy;
+    DATA dummy;
+    DATA *that = &dummy;
+    DATA *prev = 0;
     char *buffer = 0;
     char *b_fname = 0;
     char *b_temp1 = 0;
@@ -530,14 +583,17 @@ do_file(FILE *fp)
     int ok = HAVE_NOTHING;
     int marker = -1;
     int unified = 0;
+    int freed = 0;
     int old_unify = 0;
     int new_unify = 0;
+    int context = 1;
     char *s;
+#ifdef DEBUG
+    int line_no = 0;
+#endif
 
+    memset(&dummy, 0, sizeof(dummy));
     dummy.name = "";
-    dummy.ins = 0;
-    dummy.del = 0;
-    dummy.mod = 0;
 
     fixed_buffer(&buffer, fixed = length = BUFSIZ);
     fixed_buffer(&b_fname, length);
@@ -566,21 +622,38 @@ do_file(FILE *fp)
 	    else
 		break;
 	}
+	TRACE(("[%05d] %s\n", ++line_no, buffer));
 
 	/*
+	 * The lines identifying files in a context diff depend on how it was
+	 * invoked.  But after the header, each chunk begins with a line
+	 * containing 15 *'s.  Each chunk may contain a line-range with '***'
+	 * for the "before", and a line-range with '---' for the "after".  The
+	 * part of the chunk depicting the deletion may be absent, though the
+	 * edit line is present.
+	 *
 	 * The markers for unified diff are a little different from the normal
 	 * context-diff.  Also, the edit-lines in a unified diff won't have a
 	 * space in column 2.  Because of the missing space, we have to count
 	 * lines to ensure we do not confuse the marker lines.
 	 */
 	marker = -1;
-	if (match(buffer, "*** ")) {
+	if (that != &dummy && !strcmp(buffer, "***************")) {
+	    TRACE(("begin context chunk\n"));
+	    context = 2;
+	} else if (context == 2 && match(buffer, "*** ")) {
+	    context = 1;
+	} else if (context == 1 && match(buffer, "--- ")) {
+	    marker = 1;
+	    context = 0;
+	} else if (match(buffer, "*** ")) {
 	    marker = 0;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "--- ")) {
 	    marker = unified = 1;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "+++ ")) {
 	    marker = unified = 2;
-	} else if (unified == 2) {
+	} else if (unified == 2
+		   || ((old_unify + new_unify) == 0 && (*buffer == '@'))) {
 	    unified = 0;
 	    if (*buffer == '@') {
 		int old_base, new_base, old_size, new_size;
@@ -599,6 +672,29 @@ do_file(FILE *fp)
 		    unified = -1;
 		}
 	    }
+	} else if (unified == 1 && !context) {
+	    /*
+	     * If unified==1, we guessed we would find a "+++" line, but since
+	     * we are here, we did not find that.  The context check ensures
+	     * we do not mistake the "---" for a unified diff with that for
+	     * a context diff's "after" line-range.
+	     *
+	     * If we guessed wrong, then we probably found a data line with
+	     * "--" in the first two columns of the diff'd file.
+	     */
+	    unified = 0;
+	    TRACE(("Expected \"+++\" @%d:%s\n", __LINE__, buffer));
+	    if (prev != 0
+		&& prev != that
+		&& that->ins == 0
+		&& that->del == 0
+		&& strcmp(prev->name, that->name)) {
+		TRACE(("giveup on %ld/%ld %s\n", that->ins, that->del, that->name));
+		TRACE(("revert to %ld/%ld %s\n", prev->ins, prev->del, prev->name));
+		(void) delink(that);
+		that = prev;
+		that->del += 1;
+	    }
 	} else if (old_unify + new_unify) {
 	    switch (*buffer) {
 	    case '-':
@@ -622,8 +718,15 @@ do_file(FILE *fp)
 	} else {
 	    unified = 0;
 	}
-	if (marker > 0)
+
+	/*
+	 * Override the beginning of the line to simplify the case statement
+	 * below.
+	 */
+	if (marker > 0) {
+	    TRACE(("@%d, marker=%d, override %s\n", __LINE__, marker, buffer));
 	    (void) strncpy(buffer, "***", 3);
+	}
 
 	/*
 	 * Use the first character of the input line to determine its
@@ -663,7 +766,7 @@ do_file(FILE *fp)
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
-		s = do_merging(that, s);
+		s = do_merging(that, s, &freed);
 		that = new_data(s);
 		ok = begin_data(that);
 	    }
@@ -676,7 +779,7 @@ do_file(FILE *fp)
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
-		s = do_merging(that, s);
+		s = do_merging(that, s, &freed);
 		that = new_data(s);
 		ok = begin_data(that);
 	    }
@@ -726,7 +829,10 @@ do_file(FILE *fp)
 			&& !contain_any(b_fname, "*")
 			&& !edit_range(b_fname))
 		    ) {
-		    s = do_merging(that, b_fname);
+		    prev = that;
+		    s = do_merging(that, b_fname, &freed);
+		    if (freed)
+			prev = 0;
 		    that = new_data(s);
 		    ok = begin_data(that);
 		    TRACE(("after merge:%d:%s\n", ok, s));
@@ -829,16 +935,9 @@ summarize(void)
 	 * through the first path-separator, etc.
 	 */
 	if (prefix_opt >= 0) {
-	    int n, base;
-	    for (n = prefix_opt, base = 0; n > 0; n--) {
-		char *s = strchr(p->name + base, PATHSEP);
-		if (s == 0 || *++s == EOS)
-		    break;
-		base = (int) (s - p->name);
-	    }
-	    p->base = base;
-	    if (name_wide < (len - base))
-		name_wide = (len - base);
+	    /* p->base has been computed at node creation */
+	    if (name_wide < (len - p->base))
+		name_wide = (len - p->base);
 	} else {
 	    if (len < prefix_len || prefix_len < 0)
 		prefix_len = len;
@@ -918,7 +1017,7 @@ summarize(void)
 
 #ifdef HAVE_POPEN
 static char *
-is_compressed(char *name)
+is_compressed(const char *name)
 {
     char *verb = 0;
     char *result = 0;
@@ -932,7 +1031,7 @@ is_compressed(char *name)
 	verb = "bzip2 -dc %s";
     }
     if (verb != 0) {
-	result = (char *) malloc(strlen(verb) + len);
+	result = (char *) xmalloc(strlen(verb) + len);
 	sprintf(result, verb, name);
     }
     return result;
@@ -943,7 +1042,7 @@ is_compressed(char *name)
 static void
 usage(FILE *fp)
 {
-    static char *msg[] =
+    static const char *msg[] =
     {
 	"Usage: diffstat [options] [files]",
 	"",
@@ -970,6 +1069,26 @@ usage(FILE *fp)
 	fprintf(fp, "%s\n", msg[j]);
 }
 
+/* Wrapper around getopt that also parses "--help" and "--version".  
+ * argc, argv, opts, return value, and globals optarg, optind,
+ * opterr, and optopt are as in getopt().  help and version designate
+ * what should be returned if --help or --version are encountered. */
+static int
+getopt_helper(int argc, char *const argv[], const char *opts,
+	      int help, int version)
+{
+    if (optind < argc && argv[optind] != NULL) {
+	if (strcmp(argv[optind], "--help") == 0) {
+	    optind++;
+	    return help;
+	} else if (strcmp(argv[optind], "--version") == 0) {
+	    optind++;
+	    return version;
+	}
+    }
+    return getopt(argc, argv, opts);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -978,7 +1097,8 @@ main(int argc, char *argv[])
 
     max_width = 80;
 
-    while ((j = getopt(argc, argv, "ce:f:hkn:o:p:uvVw:")) != EOF) {
+    while ((j = getopt_helper(argc, argv, "ce:f:hkn:o:p:uvVw:", 'h', 'V'))
+	   != -1) {
 	switch (j) {
 	case 'c':
 	    comment_opt = "#";
@@ -1013,7 +1133,9 @@ main(int argc, char *argv[])
 	    verbose = 1;
 	    break;
 	case 'V':
+#ifndef	NO_IDENT
 	    if (!sscanf(Id, "%*s %*s %s", version))
+#endif
 		(void) strcpy(version, "?");
 	    printf("diffstat version %s\n", version);
 	    return (EXIT_SUCCESS);
@@ -1042,11 +1164,11 @@ main(int argc, char *argv[])
 		    }
 		    do_file(fp);
 		    (void) pclose(fp);
-		    free(command);
 		}
+		free(command);
 	    } else
 #endif
-	    if ((fp = fopen(name, "r")) != 0) {
+	    if ((fp = fopen(name, "rb")) != 0) {
 		if (show_progress) {
 		    (void) fprintf(stderr, "%s\n", name);
 		    (void) fflush(stderr);
